@@ -39,7 +39,7 @@ SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. */
 #   include "macos.h"
 #endif
 
-static const int    refreshInterval_InputWidget_ = 256;
+static const int    refreshInterval_InputWidget_ = 512;
 static const size_t maxUndo_InputWidget_         = 64;
 static const int    unlimitedWidth_InputWidget_  = 1000000; /* TODO: WrapText disables some functionality if maxWidth==0 */
 
@@ -188,9 +188,9 @@ enum iInputWidgetFlag {
     markWords_InputWidgetFlag        = iBit(8),
     needUpdateBuffer_InputWidgetFlag = iBit(9),
     enterKeyEnabled_InputWidgetFlag  = iBit(10),
-    enterKeyInsertsLineFeed_InputWidgetFlag
-                                     = iBit(11),
+    lineBreaksEnabled_InputWidgetFlag= iBit(11),
     needBackup_InputWidgetFlag       = iBit(12),
+    useReturnKeyBehavior_InputWidgetFlag = iBit(13),
 };
 
 /*----------------------------------------------------------------------------------------------*/
@@ -216,6 +216,7 @@ struct Impl_InputWidget {
     iArray          undoStack;
     int             font;
     iClick          click;
+    int             wheelAccum;
     int             cursorVis;
     uint32_t        timer;
     iTextBuf *      buffered; /* pre-rendered static text */
@@ -535,6 +536,7 @@ static void updateLine_InputWidget_(iInputWidget *d, iInputLine *line) {
     }
     const iTextMetrics tm = measure_WrapText(&wrapText, d->font);
     line->wrapLines.end = line->wrapLines.start + height_Rect(tm.bounds) / lineHeight_Text(d->font);
+    iAssert(!isEmpty_Range(&line->wrapLines));
 }
 
 static void updateLineRangesStartingFrom_InputWidget_(iInputWidget *d, int y) {
@@ -565,6 +567,28 @@ static void updateAllLinesAndResizeHeight_InputWidget_(iInputWidget *d) {
     }
 }
 
+static uint32_t cursorTimer_(uint32_t interval, void *w) {
+    iInputWidget *d = w;
+    if (d->cursorVis > 1) {
+        d->cursorVis--;
+    }
+    else {
+        d->cursorVis ^= 1;
+    }
+    refresh_Widget(w);
+    return interval;
+}
+
+static void startOrStopCursorTimer_InputWidget_(iInputWidget *d, iBool doStart) {
+    if (doStart && !d->timer) {
+        d->timer = SDL_AddTimer(refreshInterval_InputWidget_, cursorTimer_, d);        
+    }
+    else if (!doStart && d->timer) {
+        SDL_RemoveTimer(d->timer);
+        d->timer = 0;
+    }
+}
+
 void init_InputWidget(iInputWidget *d, size_t maxLen) {
     iWidget *w = &d->widget;
     init_Widget(w);
@@ -586,10 +610,11 @@ void init_InputWidget(iInputWidget *d, size_t maxLen) {
     d->cursor       = zero_I2();
     d->prevCursor   = zero_I2();
     d->lastUpdateWidth = 0;
-    d->inFlags         = eatEscape_InputWidgetFlag | enterKeyEnabled_InputWidgetFlag;
-    if (deviceType_App() != desktop_AppDeviceType) {
-        d->inFlags |= enterKeyInsertsLineFeed_InputWidgetFlag;
-    }
+    d->inFlags         = eatEscape_InputWidgetFlag | enterKeyEnabled_InputWidgetFlag |
+                         lineBreaksEnabled_InputWidgetFlag | useReturnKeyBehavior_InputWidgetFlag;
+    //    if (deviceType_App() != desktop_AppDeviceType) {
+    //        d->inFlags |= enterKeyInsertsLineFeed_InputWidgetFlag;
+    //    }
     iZap(d->mark);
     setMaxLen_InputWidget(d, maxLen);
     d->visWrapLines.start = 0;
@@ -599,6 +624,7 @@ void init_InputWidget(iInputWidget *d, size_t maxLen) {
     splitToLines_(&iStringLiteral(""), &d->lines);
     setFlags_Widget(w, fixedHeight_WidgetFlag, iTrue); /* resizes its own height */
     init_Click(&d->click, d, SDL_BUTTON_LEFT);
+    d->wheelAccum = 0;
     d->timer = 0;
     d->cursorVis = 0;
     d->buffered = NULL;
@@ -624,9 +650,7 @@ void deinit_InputWidget(iInputWidget *d) {
     delete_TextBuf(d->buffered);
     clearUndo_InputWidget_(d);
     deinit_Array(&d->undoStack);
-    if (d->timer) {
-        SDL_RemoveTimer(d->timer);
-    }
+    startOrStopCursorTimer_InputWidget_(d, iFalse);    
     deinit_String(&d->srcHint);
     deinit_String(&d->hint);
     deinit_String(&d->oldText);
@@ -667,7 +691,6 @@ void setMode_InputWidget(iInputWidget *d, enum iInputMode mode) {
     d->mode = mode;
 }
 
-#if 0
 static void restoreDefaultScheme_(iString *url) {
     iUrl parts;
     init_Url(&parts, url);
@@ -682,17 +705,14 @@ static const iString *omitDefaultScheme_(iString *url) {
     }
     return url;
 }
-#endif
 
 const iString *text_InputWidget(const iInputWidget *d) {
     if (d) {
         iString *text = collect_String(text_InputWidget_(d));
-#if 0
         if (d->inFlags & isUrl_InputWidgetFlag) {
             /* Add the "gemini" scheme back if one is omitted. */
             restoreDefaultScheme_(text);
         }
-#endif
         return text;
     }
     return collectNew_String();
@@ -709,10 +729,19 @@ void setMaxLen_InputWidget(iInputWidget *d, size_t maxLen) {
 }
 
 void setLineLimits_InputWidget(iInputWidget *d, int minLines, int maxLines) {
-    d->minWrapLines = minLines;
-    d->maxWrapLines = maxLines;
-    updateVisible_InputWidget_(d);
-    updateMetrics_InputWidget_(d);
+    if (d->minWrapLines != minLines || d->maxWrapLines != maxLines) {
+        d->minWrapLines = minLines;
+        d->maxWrapLines = maxLines;
+        updateVisible_InputWidget_(d);
+        updateMetrics_InputWidget_(d);
+    }
+}
+
+int minLines_InputWidget(const iInputWidget *d) {
+    return d->minWrapLines;
+}
+int maxLines_InputWidget(const iInputWidget *d) {
+    return d->maxWrapLines;
 }
 
 void setValidator_InputWidget(iInputWidget *d, iInputWidgetValidatorFunc validator, void *context) {
@@ -720,14 +749,18 @@ void setValidator_InputWidget(iInputWidget *d, iInputWidgetValidatorFunc validat
     d->validatorContext = context;
 }
 
-void setEnterInsertsLF_InputWidget(iInputWidget *d, iBool enterInsertsLF) {
-    iChangeFlags(d->inFlags, enterKeyInsertsLineFeed_InputWidgetFlag, enterInsertsLF);
+void setLineBreaksEnabled_InputWidget(iInputWidget *d, iBool lineBreaksEnabled) {
+    iChangeFlags(d->inFlags, lineBreaksEnabled_InputWidgetFlag, lineBreaksEnabled);
 }
 
 void setEnterKeyEnabled_InputWidget(iInputWidget *d, iBool enterKeyEnabled) {
     iChangeFlags(d->inFlags, enterKeyEnabled_InputWidgetFlag, enterKeyEnabled);
 }
 
+void setUseReturnKeyBehavior_InputWidget(iInputWidget *d, iBool useReturnKeyBehavior) {
+    iChangeFlags(d->inFlags, useReturnKeyBehavior_InputWidgetFlag, useReturnKeyBehavior);
+}
+
 void setHint_InputWidget(iInputWidget *d, const char *hintText) {
     /* Keep original for retranslations. */
     setCStr_String(&d->srcHint, hintText);
@@ -767,6 +800,7 @@ static iRangei visibleLineRange_InputWidget_(const iInputWidget *d) {
         }
         else break;
     }
+    iAssert(isEmpty_Range(&vis) || (vis.start >= 0 && vis.end >= vis.start));
     return vis;
 }
 
@@ -782,6 +816,22 @@ static void updateBuffered_InputWidget_(iInputWidget *d) {
         for (int i = visRange.start; i < visRange.end; i++) {
             append_String(visText, &line_InputWidget_(d, i)->text);
         }
+        if (d->inFlags & isUrl_InputWidgetFlag) {
+            /* Highlight the host name. */
+            iUrl parts;
+            init_Url(&parts, visText);
+            if (!isEmpty_Range(&parts.host)) {
+                const char *cstr = cstr_String(visText);
+                insertData_Block(&visText->chars,
+                                 parts.host.end - cstr,
+                                 restore_ColorEscape,
+                                 strlen(restore_ColorEscape));
+                insertData_Block(&visText->chars,
+                                 parts.host.start - cstr,
+                                 uiTextStrong_ColorEscape,
+                                 strlen(uiTextStrong_ColorEscape));
+            }
+        }
         iWrapText wt = wrap_InputWidget_(d, 0);
         wt.text = range_String(visText);
         const int fg = uiInputText_ColorId;
@@ -815,12 +865,10 @@ void setText_InputWidget(iInputWidget *d, const iString *text) {
             punyEncodeUrlHost_String(enc);
             text = enc;
         }
-#if 0
         /* Omit the default (Gemini) scheme if there isn't much space. */
         if (isNarrow_Root(as_Widget(d)->root)) {
             text = omitDefaultScheme_(collect_String(copy_String(text)));
         }
-#endif
     }
     clearUndo_InputWidget_(d);
     iString *nfcText = collect_String(copy_String(text));
@@ -835,10 +883,6 @@ void setText_InputWidget(iInputWidget *d, const iString *text) {
     if (!isFocused_Widget(d)) {
         iZap(d->mark);
     }
-//    else {
-//        d->cursor.y = iMin(d->cursor.y, (int) size_Array(&d->lines) - 1);
-//        d->cursor.x = iMin(d->cursor.x, size_String(&cursorLine_InputWidget_(d)->text));
-//    }
     if (!isFocused_Widget(d)) {
         d->inFlags |= needUpdateBuffer_InputWidgetFlag;
     }
@@ -853,18 +897,6 @@ void setTextCStr_InputWidget(iInputWidget *d, const char *cstr) {
     delete_String(str);
 }
 
-static uint32_t cursorTimer_(uint32_t interval, void *w) {
-    iInputWidget *d = w;
-    if (d->cursorVis > 1) {
-        d->cursorVis--;
-    }
-    else {
-        d->cursorVis ^= 1;
-    }
-    refresh_Widget(w);
-    return interval;
-}
-
 static size_t cursorToIndex_InputWidget_(const iInputWidget *d, iInt2 pos) {
     if (pos.y < 0) {
         return 0;
@@ -917,7 +949,7 @@ void begin_InputWidget(iInputWidget *d) {
     setFlags_Widget(w, selected_WidgetFlag, iTrue);
     showCursor_InputWidget_(d);
     refresh_Widget(w);
-    d->timer = SDL_AddTimer(refreshInterval_InputWidget_, cursorTimer_, d);
+    startOrStopCursorTimer_InputWidget_(d, iTrue);
     d->inFlags &= ~enterPressed_InputWidgetFlag;
     if (d->inFlags & selectAllOnFocus_InputWidgetFlag) {
         d->mark = (iRanges){ 0, lastLine_InputWidget_(d)->range.end };
@@ -942,8 +974,7 @@ void end_InputWidget(iInputWidget *d, iBool accept) {
         splitToLines_(&d->oldText, &d->lines);
     }
     d->inFlags |= needUpdateBuffer_InputWidgetFlag;
-    SDL_RemoveTimer(d->timer);
-    d->timer = 0;
+    startOrStopCursorTimer_InputWidget_(d, iFalse);
     SDL_StopTextInput();
     setFlags_Widget(w, selected_WidgetFlag | keepOnTop_WidgetFlag, iFalse);
     const char *id = cstr_String(id_Widget(as_Widget(d)));
@@ -1184,6 +1215,7 @@ static iBool deleteMarked_InputWidget_(iInputWidget *d) {
     if (!isEmpty_Range(&m)) {
         deleteIndexRange_InputWidget_(d, m);
         setCursor_InputWidget(d, indexToCursor_InputWidget_(d, m.start));
+        iZap(d->mark); /* setCursor thinks we're marking when Shift is down */
         return iTrue;
     }
     return iFalse;
@@ -1287,8 +1319,8 @@ static iInt2 coordCursor_InputWidget_(const iInputWidget *d, iInt2 coord) {
 
 static iBool copy_InputWidget_(iInputWidget *d, iBool doCut) {
     if (!isEmpty_Range(&d->mark)) {
-        const iRanges m = mark_InputWidget_(d);
-        iString *str = collectNew_String();
+        const iRanges m   = mark_InputWidget_(d);
+        iString *     str = collectNew_String();
         mergeLinesRange_(&d->lines, m, str);
         SDL_SetClipboardText(
             cstr_String(d->inFlags & isUrl_InputWidgetFlag ? canonicalUrl_String(str) : str));
@@ -1369,17 +1401,29 @@ static iBool isArrowUpDownConsumed_InputWidget_(const iInputWidget *d) {
     return d->maxWrapLines > 1;
 }
 
+static iBool checkLineBreakMods_InputWidget_(const iInputWidget *d, int mods) {
+    if (d->inFlags & useReturnKeyBehavior_InputWidgetFlag) {
+        return mods == lineBreakKeyMod_ReturnKeyBehavior(prefs_App()->returnKey);
+    }
+    return mods == 0;
+}
+
+static iBool checkAcceptMods_InputWidget_(const iInputWidget *d, int mods) {
+    if (d->inFlags & useReturnKeyBehavior_InputWidgetFlag) {
+        return mods == acceptKeyMod_ReturnKeyBehavior(prefs_App()->returnKey);
+    }
+    return mods == 0;
+}
+
 static iBool processEvent_InputWidget_(iInputWidget *d, const SDL_Event *ev) {
     iWidget *w = as_Widget(d);
     /* Resize according to width immediately. */
     if (d->lastUpdateWidth != w->rect.size.x) {
         d->inFlags |= needUpdateBuffer_InputWidgetFlag;
-#if 0
         if (d->inFlags & isUrl_InputWidgetFlag) {
             /* Restore/omit the default scheme if necessary. */
             setText_InputWidget(d, text_InputWidget(d));
         }
-#endif
         updateAllLinesAndResizeHeight_InputWidget_(d);
         d->lastUpdateWidth = w->rect.size.x;
     }
@@ -1387,6 +1431,13 @@ static iBool processEvent_InputWidget_(iInputWidget *d, const SDL_Event *ev) {
         begin_InputWidget(d);
         return iFalse;
     }
+    else if (isEditing_InputWidget_(d) && (isCommand_UserEvent(ev, "window.focus.lost") ||
+                                           isCommand_UserEvent(ev, "window.focus.gained"))) {
+        startOrStopCursorTimer_InputWidget_(d, isCommand_UserEvent(ev, "window.focus.gained"));
+        d->cursorVis = 1;
+        refresh_Widget(d);
+        return iFalse;
+    }
     else if (isCommand_UserEvent(ev, "keyroot.changed")) {
         d->inFlags |= needUpdateBuffer_InputWidgetFlag;
     }
@@ -1455,6 +1506,31 @@ static iBool processEvent_InputWidget_(iInputWidget *d, const SDL_Event *ev) {
                              ? SDL_SYSTEM_CURSOR_IBEAM
                              : SDL_SYSTEM_CURSOR_ARROW);
     }
+    if (ev->type == SDL_MOUSEWHEEL && contains_Widget(w, coord_MouseWheelEvent(&ev->wheel))) {
+        const int lineHeight = lineHeight_Text(d->font);
+        if (isPerPixel_MouseWheelEvent(&ev->wheel)) {
+            d->wheelAccum -= ev->wheel.y;
+        }
+        else {
+            d->wheelAccum -= ev->wheel.y * 3 * lineHeight;
+        }
+        int lineDelta = d->wheelAccum / lineHeight;
+        if (lineDelta < 0) {
+            lineDelta = iMax(lineDelta, -d->visWrapLines.start);
+            if (!lineDelta) d->wheelAccum = 0;
+        }
+        else if (lineDelta > 0) {
+            lineDelta = iMin(lineDelta,
+                             lastLine_InputWidget_(d)->wrapLines.end - d->visWrapLines.end);
+            if (!lineDelta) d->wheelAccum = 0;            
+        }
+        d->wheelAccum         -= lineDelta * lineHeight;
+        d->visWrapLines.start += lineDelta;
+        d->visWrapLines.end   += lineDelta;
+        d->inFlags |= needUpdateBuffer_InputWidgetFlag;
+        refresh_Widget(d);
+        return iTrue;
+    }
     switch (processEvent_Click(&d->click, ev)) {
         case none_ClickResult:
             break;
@@ -1485,6 +1561,7 @@ static iBool processEvent_InputWidget_(iInputWidget *d, const SDL_Event *ev) {
                     selectAll_InputWidget(d);
                 }
             }
+            refresh_Widget(d);
             return iTrue;
         }
         case aborted_ClickResult:
@@ -1574,10 +1651,10 @@ static iBool processEvent_InputWidget_(iInputWidget *d, const SDL_Event *ev) {
                 return iTrue;
             case SDLK_RETURN:
             case SDLK_KP_ENTER:
-                if (~d->inFlags & isSensitive_InputWidgetFlag && d->maxLen == 0) {
-                    if (mods == lineBreakKeyMod_ReturnKeyBehavior(prefs_App()->returnKey) ||
-                        (~d->inFlags & isUrl_InputWidgetFlag &&
-                         d->inFlags & enterKeyInsertsLineFeed_InputWidgetFlag)) {
+                if (~d->inFlags & isSensitive_InputWidgetFlag &&
+                    ~d->inFlags & isUrl_InputWidgetFlag &&
+                    d->inFlags & lineBreaksEnabled_InputWidgetFlag && d->maxLen == 0) {
+                    if (checkLineBreakMods_InputWidget_(d, mods)) {
                         pushUndo_InputWidget_(d);
                         deleteMarked_InputWidget_(d);
                         insertChar_InputWidget_(d, '\n');
@@ -1586,7 +1663,8 @@ static iBool processEvent_InputWidget_(iInputWidget *d, const SDL_Event *ev) {
                     }
                 }
                 if (d->inFlags & enterKeyEnabled_InputWidgetFlag &&
-                    mods == acceptKeyMod_ReturnKeyBehavior(prefs_App()->returnKey)) {
+                    (checkAcceptMods_InputWidget_(d, mods) ||
+                     (~d->inFlags & lineBreaksEnabled_InputWidgetFlag))) {
                     d->inFlags |= enterPressed_InputWidgetFlag;
                     setFocus_Widget(NULL);
                     return iTrue;
@@ -1681,13 +1759,16 @@ static iBool processEvent_InputWidget_(iInputWidget *d, const SDL_Event *ev) {
                     setCursor_InputWidget(d, key == SDLK_HOME ? zero_I2() : curMax);
                 }
                 else {
-                    setCursor_InputWidget(d, key == SDLK_HOME ? lineFirst : lineLast);
+                    moveCursorByLine_InputWidget_(d, 0, key == SDLK_HOME ? -1 : +1);
                 }
                 refresh_Widget(w);
                 return iTrue;
             case SDLK_a:
 #if defined (iPlatformApple)
                 if (mods == KMOD_PRIMARY) {
+#else
+                if (mods == (KMOD_PRIMARY | KMOD_SHIFT)) {
+#endif
                     selectAll_InputWidget(d);
                     d->mark.start = 0;
                     d->mark.end   = cursorToIndex_InputWidget_(d, curMax);
@@ -1696,7 +1777,6 @@ static iBool processEvent_InputWidget_(iInputWidget *d, const SDL_Event *ev) {
                     refresh_Widget(w);
                     return iTrue;
                 }
-#endif
                 /* fall through for Emacs-style Home/End */
             case SDLK_e:
                 if (mods == KMOD_CTRL || mods == (KMOD_CTRL | KMOD_SHIFT)) {
@@ -1857,7 +1937,7 @@ static void draw_InputWidget_(const iInputWidget *d) {
         drawRange_Text(d->font, drawPos, uiAnnotation_ColorId, range_String(&d->hint));
     }
     else {
-        /* TODO: Make a function out of this. */
+        iAssert(~d->inFlags & isSensitive_InputWidgetFlag || size_Range(&visLines) == 1);
         drawPos.y += visLineOffsetY;
         iMarkPainter marker = {
             .paint = &p,
@@ -1865,25 +1945,23 @@ static void draw_InputWidget_(const iInputWidget *d) {
             .contentBounds = contentBounds,
             .mark = mark_InputWidget_(d)
         };
-        iAssert(~d->inFlags & isSensitive_InputWidgetFlag || size_Range(&visLines) == 1);
+        wrapText.context = &marker;
+        wrapText.wrapFunc = isFocused ? draw_MarkPainter_ : NULL; /* mark is drawn under each line of text */
         for (size_t vis = visLines.start; vis < visLines.end; vis++) {
             const iInputLine *line = constAt_Array(&d->lines, vis);
-            wrapText.text     = range_String(&line->text);
-            wrapText.wrapFunc = isFocused ? draw_MarkPainter_ : NULL; /* mark is drawn under each line of text */
-            wrapText.context  = &marker;
-            marker.line       = line;
-            marker.pos        = drawPos;
+            wrapText.text = range_String(&line->text);
+            marker.line   = line;
+            marker.pos    = drawPos;
             addv_I2(&drawPos, draw_WrapText(&wrapText, d->font, drawPos, fg).advance); /* lines end with \n */
         }
         wrapText.wrapFunc = NULL;
         wrapText.context  = NULL;
     }
-    unsetClip_Paint(&p);
     /* Draw the insertion point. */
-    if (isFocused && d->cursorVis) {
-        iInt2 curSize;
-        iRangecc cursorChar = iNullRange;
-        int visWrapsAbove = 0;
+    if (isFocused && d->cursorVis && contains_Range(&visLines, d->cursor.y)) {
+        iInt2    curSize;
+        iRangecc cursorChar    = iNullRange;
+        int      visWrapsAbove = 0;
         for (int i = d->cursor.y - 1; i >= visLines.start; i--) {
             const iInputLine *line = constAt_Array(&d->lines, i);
             visWrapsAbove += numWrapLines_InputLine_(line);
@@ -1927,6 +2005,7 @@ static void draw_InputWidget_(const iInputWidget *d) {
                            cursorChar);
         }
     }
+    unsetClip_Paint(&p);
     drawChildren_Widget(w);
 }
 
