@@ -30,6 +30,8 @@ SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. */
 #include "keys.h"
 #include "touch.h"
 
+#include <SDL_version.h>
+
 struct Impl_LabelWidget {
     iWidget widget;
     iString srcLabel;
@@ -44,6 +46,9 @@ struct Impl_LabelWidget {
     struct {
         uint8_t alignVisual     : 1; /* align according to visible bounds, not font metrics */
         uint8_t noAutoMinHeight : 1; /* minimum height is not set automatically */
+        uint8_t drawAsOutline   : 1; /* draw as outline, filled with background color */
+        uint8_t noTopFrame      : 1;
+        uint8_t wrap            : 1;
     } flags;
 };
 
@@ -206,7 +211,7 @@ static void getColors_LabelWidget_(const iLabelWidget *d, int *bg, int *fg, int
         }
     }
     int colorEscape = none_ColorId;
-    if (startsWith_String(&d->label, "\r")) {
+    if (startsWith_String(&d->label, "\v")) {
         colorEscape = cstr_String(&d->label)[1] - asciiBase_ColorEscape; /* TODO: can be two bytes long */
     }
     if (isHover_LabelWidget_(d)) {
@@ -260,6 +265,14 @@ iLocalDef int iconPadding_LabelWidget_(const iLabelWidget *d) {
     return d->icon ? iRound(lineHeight_Text(d->font) * amount) : 0;
 }
 
+static iRect contentBounds_LabelWidget_(const iLabelWidget *d) {
+    iRect content = adjusted_Rect(bounds_Widget(constAs_Widget(d)),
+                                  padding_LabelWidget_(d, 0),
+                                  neg_I2(padding_LabelWidget_(d, 2)));
+    adjustEdges_Rect(&content, 0, 0, 0, iconPadding_LabelWidget_(d));
+    return content;
+}
+
 static void draw_LabelWidget_(const iLabelWidget *d) {
     const iWidget *w = constAs_Widget(d);
     draw_Widget(w);
@@ -284,12 +297,21 @@ static void draw_LabelWidget_(const iLabelWidget *d) {
         iRect frameRect = adjusted_Rect(rect, zero_I2(), init1_I2(-1));
         if (isButton) {
             iInt2 points[] = {
-                bottomLeft_Rect(frameRect), topLeft_Rect(frameRect), topRight_Rect(frameRect),
-                bottomRight_Rect(frameRect), bottomLeft_Rect(frameRect)
+                bottomLeft_Rect(frameRect),
+                topLeft_Rect(frameRect),
+                topRight_Rect(frameRect),
+                bottomRight_Rect(frameRect),
+                bottomLeft_Rect(frameRect)
             };
+#if SDL_VERSION_ATLEAST(2, 0, 16)
+            if (isOpenGLRenderer_Window()) {
+                /* A very curious regression in SDL 2.0.16. */
+                points[3].x--;    
+            }
+#endif
             drawLines_Paint(&p, points + 2, 3, frame2);
             drawLines_Paint(
-                &p, points, !isHover && flags & noTopFrame_WidgetFlag ? 2 : 3, frame);
+                &p, points, !isHover && flags & d->flags.noTopFrame ? 2 : 3, frame);
         }
     }
     setClip_Paint(&p, rect);
@@ -315,10 +337,10 @@ static void draw_LabelWidget_(const iLabelWidget *d) {
             cstr_String(&str));
         deinit_String(&str);
     }
-    if (flags & wrapText_WidgetFlag) {
-        const iRect inner = adjusted_Rect(innerBounds_Widget(w), init_I2(iconPad, 0), zero_I2());
-        const int   wrap  = inner.size.x;
-        drawWrapRange_Text(d->font, topLeft_Rect(inner), wrap, fg, range_String(&d->label));
+    if (d->flags.wrap) {
+        const iRect cont = contentBounds_LabelWidget_(d); //djusted_Rect(innerBounds_Widget(w), init_I2(iconPad, 0), zero_I2());
+        drawWrapRange_Text(
+            d->font, topLeft_Rect(cont), width_Rect(cont), fg, range_String(&d->label));
     }
     else if (flags & alignLeft_WidgetFlag) {
         draw_Text(d->font, add_I2(bounds.pos, addX_I2(padding_LabelWidget_(d, 0), iconPad)),
@@ -348,14 +370,14 @@ static void draw_LabelWidget_(const iLabelWidget *d) {
             cstr_String(&d->label));
     }
     else {
-        drawCentered_Text(d->font,
-                          adjusted_Rect(bounds,
-                                        add_I2(zero_I2(), init_I2(iconPad, 0)),
-                                        neg_I2(zero_I2())),
-                          d->flags.alignVisual,
-                          fg,
-                          "%s",
-                          cstr_String(&d->label));
+        drawCenteredOutline_Text(
+            d->font,
+            adjusted_Rect(bounds, add_I2(zero_I2(), init_I2(iconPad, 0)), neg_I2(zero_I2())),
+            d->flags.alignVisual,
+            d->flags.drawAsOutline ? fg : none_ColorId,
+            d->flags.drawAsOutline ? d->widget.bgColor : fg,
+            "%s",
+            cstr_String(&d->label));
     }
     if (flags & chevron_WidgetFlag) {
         const iRect chRect = rect;
@@ -370,12 +392,14 @@ static void draw_LabelWidget_(const iLabelWidget *d) {
 
 static void sizeChanged_LabelWidget_(iLabelWidget *d) {
     iWidget *w = as_Widget(d);
-    if (flags_Widget(w) & wrapText_WidgetFlag) {
+    if (d->flags.wrap) {
         if (flags_Widget(w) & fixedHeight_WidgetFlag) {
             /* Calculate a new height based on the wrapping. */
-            w->rect.size.y = advanceWrapRange_Text(
-                                 d->font, innerBounds_Widget(w).size.x, range_String(&d->label))
-                                 .y;
+            const iRect cont = contentBounds_LabelWidget_(d);
+            w->rect.size.y =
+                measureWrapRange_Text(d->font, width_Rect(cont), range_String(&d->label))
+                    .bounds.size.y +
+                padding_LabelWidget_(d, 0).y + padding_LabelWidget_(d, 2).y;
         }
     }
 }
@@ -383,13 +407,13 @@ static void sizeChanged_LabelWidget_(iLabelWidget *d) {
 iInt2 defaultSize_LabelWidget(const iLabelWidget *d) {
     const iWidget *w = constAs_Widget(d);
     const int64_t flags = flags_Widget(w);
-    iInt2 size = add_I2(measure_Text(d->font, cstr_String(&d->label)),
+    iInt2 size = add_I2(measure_Text(d->font, cstr_String(&d->label)).bounds.size,
                         add_I2(padding_LabelWidget_(d, 0), padding_LabelWidget_(d, 2)));
     if ((flags & drawKey_WidgetFlag) && d->key) {
         iString str;
         init_String(&str);
         keyStr_LabelWidget_(d, &str);
-        size.x += 2 * gap_UI + measure_Text(uiShortcuts_FontId, cstr_String(&str)).x;
+        size.x += 2 * gap_UI + measure_Text(uiShortcuts_FontId, cstr_String(&str)).bounds.size.x;
         deinit_String(&str);
     }
     size.x += iconPadding_LabelWidget_(d);
@@ -408,7 +432,7 @@ void updateSize_LabelWidget(iLabelWidget *d) {
         w->minSize.y = size.y; /* vertically text must remain visible */
     }
     /* Wrapped text implies that width must be defined by arrangement. */
-    if (!(flags & (fixedWidth_WidgetFlag | wrapText_WidgetFlag))) {
+    if (~flags & fixedWidth_WidgetFlag && !d->flags.wrap) {
         w->rect.size.x = size.x;
     }
     if (~flags & fixedHeight_WidgetFlag) {
@@ -440,8 +464,11 @@ void init_LabelWidget(iLabelWidget *d, const char *label, const char *cmd) {
     d->kmods = 0;
     init_Click(&d->click, d, !isEmpty_String(&d->command) ? SDL_BUTTON_LEFT : 0);
     setFlags_Widget(w, hover_WidgetFlag, d->click.button != 0);
-    d->flags.alignVisual = iFalse;
+    d->flags.alignVisual     = iFalse;
     d->flags.noAutoMinHeight = iFalse;
+    d->flags.drawAsOutline   = iFalse;
+    d->flags.noTopFrame      = iFalse;
+    d->flags.wrap            = iFalse;
     updateSize_LabelWidget(d);
     updateKey_LabelWidget_(d); /* could be bound to another key */
 }
@@ -467,6 +494,9 @@ void setTextColor_LabelWidget(iLabelWidget *d, int color) {
 void setText_LabelWidget(iLabelWidget *d, const iString *text) {
     updateText_LabelWidget(d, text);
     updateSize_LabelWidget(d);
+    if (isWrapped_LabelWidget(d)) {
+        sizeChanged_LabelWidget_(d);
+    }
 }
 
 void setAlignVisually_LabelWidget(iLabelWidget *d, iBool alignVisual) {
@@ -481,6 +511,20 @@ void setNoAutoMinHeight_LabelWidget(iLabelWidget *d, iBool noAutoMinHeight) {
     }
 }
 
+void setNoTopFrame_LabelWidget(iLabelWidget *d, iBool noTopFrame) {
+    d->flags.noTopFrame = noTopFrame;
+}
+
+void setWrap_LabelWidget(iLabelWidget *d, iBool wrap) {
+    d->flags.wrap = wrap;
+}
+
+void setOutline_LabelWidget(iLabelWidget *d, iBool drawAsOutline) {
+    if (d) {
+        d->flags.drawAsOutline = drawAsOutline;
+    }
+}
+
 void updateText_LabelWidget(iLabelWidget *d, const iString *text) {
     set_String(&d->label, text);
     set_String(&d->srcLabel, text);
@@ -495,6 +539,11 @@ void updateTextCStr_LabelWidget(iLabelWidget *d, const char *text) {
     refresh_Widget(&d->widget);
 }
 
+void updateTextAndResizeWidthCStr_LabelWidget(iLabelWidget *d, const char *text) {
+    updateTextCStr_LabelWidget(d, text);
+    d->widget.rect.size.x = defaultSize_LabelWidget(d).x;
+}
+
 void setTextCStr_LabelWidget(iLabelWidget *d, const char *text) {
     setCStr_String(&d->label, text);
     set_String(&d->srcLabel, &d->label);
@@ -537,6 +586,10 @@ iChar icon_LabelWidget(const iLabelWidget *d) {
     return d->icon;
 }
 
+iBool isWrapped_LabelWidget(const iLabelWidget *d) {
+    return d->flags.wrap;
+}
+
 const iString *text_LabelWidget(const iLabelWidget *d) {
     if (!d) return collectNew_String();
     return &d->label;
