@@ -65,14 +65,12 @@ static float initialUiScale_ = 1.0f;
 static float initialUiScale_ = 1.1f;
 #endif
 
+static iBool isOpenGLRenderer_;
+
 iDefineTypeConstructionArgs(Window, (iRect rect), rect)
 
 /* TODO: Define menus per platform. */
 
-#if defined (iPlatformAppleDesktop)
-#  define iHaveNativeMenus
-#endif
-
 #if defined (iHaveNativeMenus)
 /* Using native menus. */
 static const iMenuItem fileMenuItems_[] = {
@@ -202,7 +200,7 @@ static void windowSizeChanged_Window_(iWindow *d) {
 }
 
 static void setupUserInterface_Window(iWindow *d) {
-#if defined (iPlatformAppleDesktop)
+#if defined (iHaveNativeMenus)
     insertMacMenus_();
 #endif
     /* One root is created by default. */
@@ -418,7 +416,7 @@ void init_Window(iWindow *d, iRect rect) {
     d->isExposed = iFalse;
     d->isMinimized = iFalse;
     d->isInvalidated = iFalse; /* set when posting event, to avoid repeated events */
-    d->isMouseInside = iTrue;
+    d->isMouseInside = iTrue;    
     d->ignoreClick = iFalse;
     d->focusGainedAt = 0;
     d->keyboardHeight = 0;
@@ -450,6 +448,7 @@ void init_Window(iWindow *d, iRect rect) {
     /* Some info. */ {
         SDL_RendererInfo info;
         SDL_GetRendererInfo(d->render, &info);
+        isOpenGLRenderer_ = !iCmpStr(info.name, "opengl");
         printf("[window] renderer: %s%s\n", info.name,
                info.flags & SDL_RENDERER_ACCELERATED ? " (accelerated)" : "");
 #if !defined (NDEBUG)
@@ -744,6 +743,7 @@ static iBool handleWindowEvent_Window_(iWindow *d, const SDL_WindowEvent *ev) {
             postRefresh_App();
             return iTrue;
         case SDL_WINDOWEVENT_RESTORED:
+        case SDL_WINDOWEVENT_SHOWN:
             updateSize_Window_(d, iTrue);
             invalidate_Window_(d, iTrue);
             d->isMinimized = iFalse;
@@ -820,7 +820,7 @@ iBool processEvent_Window(iWindow *d, const SDL_Event *ev) {
         }
         case SDL_RENDER_TARGETS_RESET:
         case SDL_RENDER_DEVICE_RESET: {
-            invalidate_Window(d);
+            invalidate_Window_(d, iTrue /* force full reset */);
             break;
         }
         default: {
@@ -913,7 +913,7 @@ iBool processEvent_Window(iWindow *d, const SDL_Event *ev) {
                 }
             }
             if (isCommand_UserEvent(&event, "lang.changed")) {
-#if defined (iPlatformAppleDesktop)
+#if defined (iHaveNativeMenus)
                 /* Retranslate the menus. */
                 removeMacMenus_();
                 insertMacMenus_();
@@ -961,10 +961,14 @@ iBool dispatchEvent_Window(iWindow *d, const SDL_Event *ev) {
             if (isCommand_SDLEvent(ev) && ev->user.data2 && ev->user.data2 != root) {
                 continue; /* Not meant for this root. */
             }
-            else if ((ev->type == SDL_KEYDOWN || ev->type == SDL_KEYUP || ev->type == SDL_TEXTINPUT)
+            if ((ev->type == SDL_KEYDOWN || ev->type == SDL_KEYUP || ev->type == SDL_TEXTINPUT)
                      && d->keyRoot != root) {
                 continue; /* Key events go only to the root with keyboard focus. */
             }
+            if (ev->type == SDL_MOUSEWHEEL && !contains_Rect(rect_Root(root),
+                                                             coord_MouseWheelEvent(&ev->wheel))) {
+                continue; /* Only process the event in the relevant split. */
+            }
             setCurrent_Root(root);
             const iBool wasUsed = dispatchEvent_Widget(root->widget, ev);
             if (wasUsed) {
@@ -1158,20 +1162,19 @@ iInt2 coord_Window(const iWindow *d, int x, int y) {
     return mulf_I2(init_I2(x, y), d->pixelRatio);
 }
 
-iInt2 mouseCoord_Window(const iWindow *d) {
-#if defined (iPlatformMobile)
-    /* At least on iOS the coordinates returned by SDL_GetMouseState() do no match up with
-       our touch coordinates on the Y axis (?). Maybe a pixel ratio thing? */
-    iUnused(d);
-    return latestPosition_Touch();
-#else
+iInt2 mouseCoord_Window(const iWindow *d, int whichDevice) {
+    if (whichDevice == SDL_TOUCH_MOUSEID) {
+        /* At least on iOS the coordinates returned by SDL_GetMouseState() do no match up with
+           our touch coordinates on the Y axis (?). Maybe a pixel ratio thing? */
+        iUnused(d);
+        return latestPosition_Touch();
+    }
     if (!d->isMouseInside) {
         return init_I2(-1000000, -1000000);
     }
     int x, y;
     SDL_GetMouseState(&x, &y);
     return coord_Window(d, x, y);
-#endif
 }
 
 float uiScale_Window(const iWindow *d) {
@@ -1187,6 +1190,10 @@ iWindow *get_Window(void) {
     return theWindow_;
 }
 
+iBool isOpenGLRenderer_Window(void) {
+    return isOpenGLRenderer_;
+}
+
 void setKeyboardHeight_Window(iWindow *d, int height) {
     if (d->keyboardHeight != height) {
         d->keyboardHeight = height;
